from datetime import date
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-tfhub-yamnet']):
    import tensorflow as tf
    import tensorflow_hub as hub

from ..base import BaseAudio2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition

YamnetModelDefinition = ModelDefinition(markdown_filepath='encoders/audio/tfhub/yamnet')
__doc__ = YamnetModelDefinition.create_docs()

class Yamnet2Vec(BaseAudio2Vec):
    definition = YamnetModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/yamnet/1'):
        self.model_url = model_url
        self.model = hub.load(self.model_url)
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 1024
    
    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/yamnet/1': {'vector_length': 1024}
        }

    @catch_vector_errors
    def encode(self, audio, vector_operation='mean', layer='embeddings'):
        if isinstance(audio, str):
            audio = self.read(audio)
        outputs = self.model(audio)
        if layer == 'scores':
            return self._vector_operation(outputs[0], vector_operation)
        elif layer == 'log_mel_spectrogram':
            return self._vector_operation(outputs[2], vector_operation)
        else:
            return self._vector_operation(outputs[1], vector_operation)
    
    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean', layer='embeddings'):
        audios = [self.read(audio) if isinstance(audio, str) else audio for audio in audios]
        return [self.encode(audio, vector_operation=vector_operation, layer=layer) for audio in audios]



from ...import_utils import *

if is_all_dependency_installed('audio-encoder'):
    import librosa
    import soundfile as sf

import tempfile
import shutil
import os
from urllib.request import urlopen, Request
from urllib.parse import quote
import io
import numpy as np

from ...base import Base2Vec, catch_vector_errors

class BaseAudio2Vec(Base2Vec):
    def read(self, audio: str, new_sampling_rate: int = 16000):
        """An method to specify the read method to read the data.
        """
        if type(audio) == str:
            if 'http' in audio:
                fd, fp = tempfile.mkstemp()
                os.write(fd, urlopen(Request(quote(audio, safe=':/?*=\''),
                                             headers={'User-Agent': "Magic Browser"})).read())
                if '.mp3' in audio:
                    data, sampling_rate = librosa.load(fp, dtype='float32')
                else:
                    data, sampling_rate = sf.read(fp, dtype='float32')
                os.close(fd)
            else:
                data, sampling_rate = sf.read(audio, dtype='float32')
        elif type(audio) == bytes:
            data, sampling_rate = sf.read(io.BytesIO(audio), dtype='float32')
        elif type(audio) == io.BytesIO:
            data, sampling_rate = sf.read(audio, dtype='float32')
        return np.array(librosa.resample(data.T, sampling_rate, new_sampling_rate))
    
    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        return [self.encode(c, vector_operation) for c in audios]



from datetime import date
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ..base import BaseAudio2Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-tfhub-vggish']):
    import tensorflow as tf
    import tensorflow_hub as hub


VggishModelDefinition = ModelDefinition(markdown_filepath='encoders/audio/tfhub/vggish')

__doc__ = VggishModelDefinition.create_docs()

class Vggish2Vec(BaseAudio2Vec):
    definition = VggishModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/vggish/1'):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '')
        self.model = hub.load(self.model_url)
        self.vector_length = 128
    
    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/vggish/1': {'vector_length': 128}
        }

    @catch_vector_errors
    def encode(self, audio, vector_operation='mean'):
        if isinstance(audio, str):
            audio = self.read(audio)
        return self._vector_operation(self.model(audio), vector_operation)

    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        audios = [self.read(audio) if isinstance(audio, str) else audio for audio in audios]
        return [self.encode(audio, vector_operation=vector_operation) for audio in audios]



from datetime import date
from ....import_utils import *
from ....models_dict import *
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-tfhub-trill']):
    import tensorflow as tf
    import tensorflow_hub as hub
    import traceback
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ..base import BaseAudio2Vec

TrillModelDefinition = ModelDefinition(markdown_filepath='encoders/audio/tfhub/trill')
__doc__ = TrillModelDefinition.create_docs()

class Trill2Vec(BaseAudio2Vec):
    definition = TrillModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/nonsemantic-speech-benchmark/trill/3', 
    layer: str = 'embedding'):
        self.model_url = model_url
        self.layer = layer
        self.model = hub.load(self.model_url)
        self.model_name = model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 512

    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/nonsemantic-speech-benchmark/trill/3': {'vector_length': 512}
        }
    
    @catch_vector_errors
    def encode(self, audio, vector_operation='mean'):
        """
        
        Example:
            >>> from encoders.audio.trill import Trill2Vec
            >>> encoder = Trill2Vec()
            >>> encoder.encode(...)
        
        """
        if isinstance(audio, str):
            audio = self.read(audio)
        return self._vector_operation(self.model(samples=audio, sample_rate=16000)[self.layer], vector_operation)

    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        audios = [self.read(audio) if isinstance(audio, str) else audio for audio in audios]
        return [self.encode(audio) for audio in audios]



from typing import List, Union
from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-torch-transformers-auto']):
    from transformers import LongformerTokenizer, LongformerModel
    import torch

LongformerModelDefinition = ModelDefinition(markdown_filepath='encoders/text/torch_transformers/torch_longformers.md')

__doc__ = LongformerModelDefinition.create_docs()

class Longformer2Vec(BaseText2Vec):
    definition = LongformerModelDefinition

    def __init__(self, model_name: str = "allenai/longformer-base-4096"):
        self.model = LongformerModel.from_pretrained(model_name)
        self.tokenizer = LongformerTokenizer.from_pretrained(model_name)

    @staticmethod
    def list_possible_models():
        return {
            'allenai/longformer-base-4096': 'Starting from RoBERTa-base checkpoint, trained on documents of max length 4,096',
            'allenai/longformer-large-4096': 'Starting from RoBERTa-large checkpoint, trained on documents of max length 4,096'
        }

    @property
    def urls(self):
        return {
            'allenai/longformer-base-4096': {'vector_length': 4096},
            'allenai/longformer-large-4096': {'vector_length': 4096}
        }

    @catch_vector_errors
    def encode(self, text: Union[str, List[str]]) -> List[float]:
        """
            Encode words using transformers.
            Args:
                text: str

        """
        if isinstance(text, str):
            return torch.mean(self.model(**self.tokenizer(text, return_tensors='pt'))[0], axis=1).detach().tolist()[0]
        if isinstance(text, list):
            return self.bulk_encode(text)
        raise ValueError(
            "Not a string or a list of strings, please enter valid data type.")

    @catch_vector_errors
    def bulk_encode(self, texts: List[str], pooling_method='mean') -> List[List[float]]:
        """
            Encode multiple sentences using transformers.
            args:
                texts: List[str]
        """
        # We use pad_to_multiple_of as other arguments usually do not work.
        if pooling_method == 'mean':
            return torch.mean(
                self.model(**self.tokenizer(texts, return_tensors='pt', 
                pad_to_multiple_of=self.tokenizer.model_max_length,
                truncation=True, padding=True))[0], axis=1).detach().tolist()
        else:
            raise NotImplementedError



"""
    Various utilities for VectorHub.
"""
import json
import os
from pathlib import Path
from pkg_resources import resource_filename
from collections import defaultdict
from .models_dict import *
from .import_utils import *

def list_installed_models(extra_requirements_file: str=resource_filename('vectorhub', '../extra_requirements.json')):
    """
        List models that are installed.
        We use resource_filename to resolve relative directory issues.
    """
    requirements = json.load(open(extra_requirements_file))
    print("The following packages are available to be used: ")
    all_available_models = []
    for package, dependency in MODEL_REQUIREMENTS.items():
        if is_all_dependency_installed(dependency, raise_warning=False):
            print(package)
            all_available_models.append(package)
    return all_available_models

def list_models(return_names_only=False):
    """
        List available models.
        Args:
            return_names_only: Return the model names
    """
    if return_names_only:
        return [x.stem for x in list(Path('.').glob('**/[!_]*.py'))]
    
    all_models = [str(x).replace('.py', '') for x in list(Path(resource_filename('vectorhub', 'encoders/text')).rglob(f'[!_]*.py'))] + \
        [str(x).replace('.py', '') for x in list(Path(resource_filename('vectorhub', 'encoders/image')).rglob(f'[!_]*.py'))] + \
        [str(x).replace('.py', '') for x in list(Path(resource_filename('vectorhub', 'encoders/audio')).rglob(f'[!_]*.py'))] + \
        [str(x).replace('.py', '') for x in list(Path(resource_filename('vectorhub', 'bi_encoders/text_text')).rglob(f'[!_]*.py'))]
    return [x.split('vectorhub/')[-1] for x in all_models if '/base' not in x]




from ..base import BaseTextText2Vec
from ....doc_utils import ModelDefinition
from ....import_utils import *
from datetime import date
if is_all_dependency_installed('encoders-text-torch-transformers'):
    from transformers import DPRContextEncoder, DPRContextEncoderTokenizer, DPRQuestionEncoder, DPRQuestionEncoderTokenizer, DPRReader, DPRReaderTokenizer
    import torch
    import numpy as np

DPRModelDefinition = ModelDefinition(markdown_filepath='bi_encoders/text_text/torch_transformers/dpr')
__doc__ = DPRModelDefinition.create_docs()

class DPR2Vec(BaseTextText2Vec):
    definition = DPRModelDefinition
    def __init__(self):
        self.context_tokenizer = DPRContextEncoderTokenizer.from_pretrained('facebook/dpr-ctx_encoder-single-nq-base')
        self.context_model = DPRContextEncoder.from_pretrained('facebook/dpr-ctx_encoder-single-nq-base', return_dict=True)

        self.query_tokenizer = DPRQuestionEncoderTokenizer.from_pretrained('facebook/dpr-question_encoder-single-nq-base')
        self.query_encoder = DPRQuestionEncoder.from_pretrained("facebook/dpr-question_encoder-single-nq-base")

        self.reader_tokenizer = DPRReaderTokenizer.from_pretrained('facebook/dpr-reader-single-nq-base')
        self.reader_model = DPRReader.from_pretrained('facebook/dpr-reader-single-nq-base', return_dict=True)
        self.vector_length = 768

    def encode_question(self, question):
        """
            Encode a question with DPR.
        """
        input_ids = self.query_tokenizer(question, return_tensors='pt')["input_ids"]
        return self.query_encoder(input_ids)[0].tolist()[0]

    def bulk_encode_questions(self, questions: str):
        """
            Bulk encode the question
        """
        input_ids = self.query_tokenizer(questions, truncation=True, max_length=True, return_tensors='pt')["input_ids"]
        return self.query_encoder(input_ids)[0].tolist()

    def encode_answer(self, answer: str):
        """
            Encode an answer with DPR.
        """
        if isinstance(answer, str):
            input_ids = self.context_tokenizer(answer, return_tensors='pt', truncation=True, 
            max_length=512)["input_ids"]
            return self.context_model(input_ids).pooler_output.tolist()[0]
        elif isinstance(answer, list):
            return self.bulk_encode_answers(answer)

    def bulk_encode_answers(self, answers: str):
        """
            Bulk encode the answers with DPR.
        """
        input_ids = self.context_tokenizer(answers, return_tensors='pt', truncation=True, padding=True,
        max_length=512)["input_ids"]
        return self.context_model(input_ids).pooler_output.tolist()



from datetime import date
from ....import_utils import *
from ....models_dict import *
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-tfhub-trill']):
    import tensorflow as tf
    import tensorflow_hub as hub
    import traceback
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ..base import BaseAudio2Vec

TrillDistilledModelDefinition = ModelDefinition(markdown_filepath="encoders/audio/tfhub/trill_distilled")
__doc__ = TrillDistilledModelDefinition.create_docs()

class TrillDistilled2Vec(BaseAudio2Vec):
    definition = TrillDistilledModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/nonsemantic-speech-benchmark/trill-distilled/3', layer: str = 'embedding'):
        self.model_url = model_url
        self.layer = layer
        self.model = hub.load(self.model_url)
        self.model_name = model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 2048

    @property
    def urls(self):
        """
        Return the URLS of models and their vector length
        """
        return {
            'https://tfhub.dev/google/nonsemantic-speech-benchmark/trill-distilled/3': {'vector_length': 2048}
        }


    @catch_vector_errors
    def encode(self, audio, vector_operation='mean', sample_rate=16000):
        if isinstance(audio, str):
            audio = self.read(audio)
        return self._vector_operation(self.model(samples=audio, sample_rate=sample_rate)[self.layer], vector_operation)
    
    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        audios = [self.read(audio) if isinstance(audio, str) else audio for audio in audios]
        return [self.encode(audio, vector_operation=vector_operation) for audio in audios]



from typing import List
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ..base import BaseImage2Vec
if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub
    import traceback

InceptionResnetModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/inception_resnet')

__doc__ = InceptionResnetModelDefinition.create_docs()

class InceptionResnet2Vec(BaseImage2Vec):
    definition = InceptionResnetModelDefinition
    def __init__(self, model_url="https://tfhub.dev/google/imagenet/inception_resnet_v2/feature_vector/4"):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.model = hub.load(self.model_url)
        self.vector_length = 1536

    @property
    def urls(self):
        return {
            "https://tfhub.dev/google/imagenet/inception_resnet_v2/feature_vector/4": {"vector_length": 1536}
        }

    @catch_vector_errors
    def encode(self, image):
        """
        Encode an image using InceptionResnet.

        Example:
            >>> from vectorhub.image.encoder.tfhub import inception_resnet
            >>> model = InceptionResnet2Vec(username, api_key)
            >>> model.encode("Hey!")
        """
        if isinstance(image, str):
            image = self.read(image)
        return self.model([image]).numpy().tolist()[0]
    
    @catch_vector_errors
    def bulk_encode(self, images):
        return [self.encode(x) for x in images]



"""
FastAI Resnet model
"""

from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ..base import BaseImage2Vec
from .base import FastAIBase

if is_all_dependency_installed('encoders-image-fastai'):
    from fastai.vision.all import (resnet18, resnet34, resnet50, resnet101, resnet152,
        squeezenet1_0, squeezenet1_1, densenet121, vgg16_bn, alexnet)

FastAIResnetModelDefinition = ModelDefinition(markdown_filepath="encoders/image/fastai/resnet_fastai")
__doc__ = FastAIResnetModelDefinition.create_docs()

class FastAIResnet2Vec(FastAIBase):
    definition = FastAIResnetModelDefinition
    def __init__(self, architecture='resnet34', databunch=None):
        """
            Refer to possible_architectures method for reference to which architectures can be instantiated. 
            Args:
                Architecture: The name of the architecture 
                Databunch: A FastAI Data collection data type that is used to instantiate a learner object.
        """
        self.databunch = databunch
        self.architecture = self.architecture_mappings[architecture]
        self._create_learner()
    
    @property
    def possible_architectures(self):
        return list(self.architecture_mappings.keys())

    @property
    def architecture_mappings(self):
        """
            Architecture mappings
        """
        return {
            'resnet18': resnet18,
            'resnet34': resnet34,
            'resnet50': resnet50,
            'resnet101': resnet101,
            'resnet152': resnet152,
            'squeezenet1_0': squeezenet1_0,
            'squeezenet1_1': squeezenet1_1,
            'densenet121': densenet121,
            'vgg16_bn': vgg16_bn,
            'alexnet': alexnet
        }

    @property
    def extraction_layer(self):
        """
            Here we selected the default to be layer_num 1 to extract the layer with the highest number of dimensions
            after it has been flattened.
        """
        return [self.learn.model[1][1]]



from typing import List, Union
from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from datetime import date
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-torch-transformers-auto']):
    from transformers import AutoTokenizer, AutoModel
    import torch

LegalBertModelDefinition = ModelDefinition(markdown_filepath='encoders/text/torch_transformers/legal_bert')

__doc__ = LegalBertModelDefinition.create_docs()

class LegalBert2Vec(BaseText2Vec):
    definition = LegalBertModelDefinition
    def __init__(self, model_name: str="nlpaueb/legal-bert-base-uncased"):
        self.model = AutoModel.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

    @staticmethod
    def list_possible_models():
        return {
            "nlpaueb/bert-base-uncased-contracts": "Trained on US contracts",
            "nlpaueb/bert-base-uncased-eurlex": "Trained on EU legislation", 
            "nlpaueb/bert-base-uncased-echr	": "Trained on ECHR cases",
            "nlpaueb/legal-bert-base-uncased": "Trained on all the above", 
            "nlpaueb/legal-bert-small-uncased": "Trained on all the above"
        }
    @property
    def urls(self):
        return {
            "nlpaueb/bert-base-uncased-contracts": {},
            "nlpaueb/bert-base-uncased-eurlex": {},
            "nlpaueb/bert-base-uncased-echr	": {},
            "nlpaueb/legal-bert-base-uncased": {},
            "nlpaueb/legal-bert-small-uncased": {},
        }

    @catch_vector_errors
    def encode(self, text: Union[str, List[str]]) -> List[float]:
        """
            Encode words using transformers.
            Args:
                text: str
        """
        if isinstance(text, str):
            return torch.mean(self.model(**self.tokenizer(text, return_tensors='pt'))[0], axis=1).detach().tolist()[0]
        if isinstance(text, list):
            return self.bulk_encode(text)
        raise ValueError("Not a string or a list of strings, please enter valid data type.")

    @catch_vector_errors
    def bulk_encode(self, texts: List[str]) -> List[List[float]]:
        """
            Encode multiple sentences using transformers.
            args:
                texts: List[str]
        """
        # We use pad_to_multiple_of as other arguments usually do not work.
        return torch.mean(self.model(**self.tokenizer(texts, return_tensors='pt', pad_to_multiple_of=self.tokenizer.model_max_length, 
        truncation=True, padding=True))[0], axis=1).detach().tolist()



import yaml
import sys
import re
import os
from datetime import date
from pkg_resources import resource_exists, resource_filename

class ModelDefinition:
    def __init__(self, model_id: str='', model_name: str='', vector_length: int='',
    description: str='', paper: str='', repo: str='', architecture: str='Not stated.',
    tasks: str='Not stated.', release_date: date='', limitations: str='Not stated.', installation: str='Not stated.',
    example: str='Not stated.', markdown_filepath: str='', **kwargs):
        """
            Model definition.
            Args:
                model_id: the identity of the model. Required for AutoEncoder.
                model_name: The name of the model
                vector_length: The length of the vector
                description: The description of the encoder
                paper: The paper which dictates the encoder
                repo: The repository fo the model
                architecture: The architecture of the model.
                task: The downstream task that the model was trained on
                limitations: The limitations of the encoder
                installation: How to isntall the encoder.
                example: The example of the encoder
        """
        self.model_id = model_id
        self.model_name = model_name
        self.vector_length = vector_length
        self.description = description
        self.paper = paper
        self.repo = repo
        self.architecture = architecture
        self.tasks = tasks
        self.release_date = release_date.__str__() if release_date is not None else None
        self.limitations = limitations
        self.installation = installation
        self.example = example
        self.markdown_filepath = markdown_filepath
        for k, v in kwargs.items():
            # assert( k in self.__class__.__allowed )
            setattr(self, k, v)
        if markdown_filepath != '':
            self.from_markdown(markdown_filepath)


    def create_docs(self):
        """
            Return a string with the RST documentation of the model.
        """
        return f"""
**Model Name**: {self.model_name}

**Vector Length**: {self.vector_length}

**Description**: {self.description}

**Paper**: {self.paper}

**Repository**: {self.repo}

**Architecture**: {self.architecture}

**Tasks**: {self.tasks}

**Release Date**: {self.release_date}

**Limitations**: {self.limitations}

**Installation**: ``{self.installation}``

**Example**:

.. code-block:: python

    {self.example}
        """

    def to_dict(self, return_base_dictionary=False):
        """
            Create a dictionary with all the attributes of the model.
        """
        if return_base_dictionary:
            return {
                "model_id": self.model_id,
                "model_name": self.model_name,
                "vector_length": self.vector_length,
                "description": self.description,
                "paper": self.paper,
                "repo": self.repo,
                "architecture": self.architecture,
                "tasks": self.tasks,
                "limitations": self.limitations,
                "installation" : self.installation,
                "example" : self.example,
                "release_date": self.release_date
            }
        else:
            model_dict = {}
            for attr in dir(self):
                if '__' in attr:
                    continue
                if isinstance(getattr(self, attr), (float, str, int)):
                    model_dict[attr] = getattr(self, attr)
            # Enforce string typecast on vector length
            model_dict['vector_length'] = str(model_dict['vector_length'])
            return model_dict

    def _get_yaml(self, f):
        """
            Returns YAML file from Python
            Args:
                f: Get YAML file.
        """
        pointer = f.tell()
        if f.readline() != '---\n':
            f.seek(pointer)
            return ''
        readline = iter(f.readline, '')
        readline = iter(readline.__next__, '---\n') #underscores needed for Python3?
        return ''.join(readline)

    def from_markdown(self, markdown_filepath: str, encoding='UTF-8', splitter=r"(\#\#+\ +)|(\n)",
    verbose=False):
        """
            Reads definitions from the markdown.
            Args:
                markdown_filepath: The path of the markdown file.
                encoding: The encoding used to open the Markdown file
        """
        if '.md' not in markdown_filepath:
            markdown_filepath += '.md'
        # Check filepath exists with
        if not os.path.exists(markdown_filepath):
            if resource_exists('vectorhub', markdown_filepath):
                markdown_filepath = resource_filename('vectorhub', markdown_filepath)
            else:
                raise FileNotFoundError(f"Unable to find {markdown_filepath}.")
        if verbose:
            print(markdown_filepath)
        # Remove sys.argv, not sure what it was doing
        with open(markdown_filepath, encoding=encoding) as f:
            config = list(yaml.load_all(self._get_yaml(f), Loader=yaml.SafeLoader))
            text = f.read()
            self.config = config[0]
            for k,v in self.config.items():
                setattr(self, k, v)
            self.markdown_description = text
        self._split_markdown_description(text, splitter=splitter)

    def _split_markdown_description(self, description: str, splitter: str=r"(\#\#+\ +)|(\n)"):
        """
            Breaks markdown into heading and values.
            Args:
                description: Description of the markdown
                splitter: Regex to split the sentence. Currently it splits on headings and new lines.
                The purpose of this is to allow us to get keys from markdown files.
        """
        # Loops through split markdown
        # If ## is detected inside string, marks the next
        # string as heading
        # and whatever follows as the value
        IS_HEADING = False
        value = ''
        heading = None
        SKIP_NEW_LINE = False
        markdown_values = {}
        for x in re.split(splitter, description):
            if x is None:
                continue
            if SKIP_NEW_LINE:
                if x == '\n':
                    continue

            if IS_HEADING:
                heading = x.lower().rstrip().replace(' ', '_')
                IS_HEADING = False
                # Skip new line after the heading is declared
                SKIP_NEW_LINE = True
                value = ""
            elif '##' in x:
                # Insert setting new layer
                if heading is not None:
                    setattr(self, heading, value)
                    markdown_values[heading] = value
                IS_HEADING = True
            else:
                SKIP_NEW_LINE = False
                value += x

        # Set the final value
        if hasattr(self, heading):
            if getattr(self, heading) != value:
                setattr(self, heading, value)
        else:
            setattr(self, heading, value)




from datetime import date
from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-labse']):
    import tensorflow as tf
    if hasattr(tf, 'executing_eagerly'):
        if not tf.executing_eagerly():
            tf.compat.v1.enable_eager_execution()
    import tensorflow_hub as hub
    import bert
    import numpy as np

LABSEModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/labse.md')
__doc__ = LABSEModelDefinition.create_docs()


class LaBSE2Vec(BaseText2Vec):
    definition = LABSEModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/LaBSE/1', max_seq_length: int = 128, normalize: bool = True):
        self.max_seq_length = max_seq_length
        self.normalize = normalize
        self.model = self.init(model_url)
        self.tokenizer = self.init_tokenizer()
        self.vector_length = 768

    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/LaBSE/1': {'vector_length': 768}
        }

    def init(self, model_url: str):
        self.model_layer = hub.KerasLayer(model_url)
        input_word_ids = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32,
                                            name="input_word_ids")
        input_mask = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32,
                                        name="input_mask")
        segment_ids = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32,
                                            name="segment_ids")
        pooled_output,  _ = self.model_layer(
            [input_word_ids, input_mask, segment_ids])
        if(self.normalize):
            pooled_output = tf.keras.layers.Lambda(
                lambda x: tf.nn.l2_normalize(x, axis=1))(pooled_output)
        return tf.keras.Model(
            inputs=[input_word_ids, input_mask, segment_ids],
            outputs=pooled_output)

    def init_tokenizer(self):
        self.vocab_file = self.model_layer.resolved_object.vocab_file.asset_path.numpy()
        self.do_lower_case = self.model_layer.resolved_object.do_lower_case.numpy()
        return bert.bert_tokenization.FullTokenizer(
            self.vocab_file, self.do_lower_case)

    def process(self, input_strings: str):
        input_ids_all, input_mask_all, segment_ids_all = [], [], []
        if isinstance(input_strings, str):
            input_strings = [input_strings]
        for input_string in input_strings:
            # Tokenize input.
            input_tokens = ["[CLS]"] + \
                self.tokenizer.tokenize(input_string) + ["[SEP]"]
            input_ids = self.tokenizer.convert_tokens_to_ids(input_tokens)
            sequence_length = min(len(input_ids), self.max_seq_length)

            # Padding or truncation.
            if len(input_ids) >= self.max_seq_length:
                input_ids = input_ids[:self.max_seq_length]
            else:
                input_ids = input_ids + [0] * \
                    (self.max_seq_length - len(input_ids))

            input_mask = [1] * sequence_length + [0] * \
                (self.max_seq_length - sequence_length)

            input_ids_all.append(input_ids)
            input_mask_all.append(input_mask)
            segment_ids_all.append([0] * self.max_seq_length)

        return np.array(input_ids_all), np.array(input_mask_all), np.array(segment_ids_all)

    @catch_vector_errors
    def encode(self, text: str):
        input_ids, input_mask, segment_ids = self.process(text)
        return self.model([input_ids, input_mask, segment_ids]).numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, texts: list):
        input_ids, input_mask, segment_ids = self.process(texts)
        return self.model([input_ids, input_mask, segment_ids]).numpy().tolist()


from typing import List
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseText2Vec
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-sentence-transformers']):
    from sentence_transformers import SentenceTransformer

SentenceTransformerModelDefinition = ModelDefinition(markdown_filepath='encoders/text/sentence_transformers/sentence_auto_transformers.md')

LIST_OF_URLS = {
    'distilroberta-base-paraphrase-v1' : {"vector_length": 768},
    'xlm-r-distilroberta-base-paraphrase-v1' : {"vector_length": 768},

    'distilbert-base-nli-stsb-mean-tokens' : {"vector_length": 768},
    'bert-large-nli-stsb-mean-tokens' : {"vector_length": 1024},
    'roberta-base-nli-stsb-mean-tokens' : {"vector_length": 768},
    'roberta-large-nli-stsb-mean-tokens' : {"vector_length": 1024},

    'distilbert-base-nli-stsb-quora-ranking' : {"vector_length": 768},
    'distilbert-multilingual-nli-stsb-quora-ranking' : {"vector_length": 768},

    'distilroberta-base-msmarco-v1' : {"vector_length": 768},

    'distiluse-base-multilingual-cased-v2' : {"vector_length": 512},
    'xlm-r-bert-base-nli-stsb-mean-tokens' : {"vector_length": 768},

    'bert-base-wikipedia-sections-mean-tokens' : {"vector_length": 768},

    'LaBSE' : {"vector_length": 768},

    'average_word_embeddings_glove.6B.300d' : {"vector_length": 300},
    'average_word_embeddings_komninos' : {"vector_length": 300},
    'average_word_embeddings_levy_dependency' : {"vector_length": 768},
    'average_word_embeddings_glove.840B.300d' : {"vector_length": 300}
}

__doc__ = SentenceTransformerModelDefinition.create_docs()


class SentenceTransformer2Vec(BaseText2Vec):
    definition = SentenceTransformerModelDefinition
    def __init__(self, model_name: str):
        self.list_of_urls = LIST_OF_URLS
        self.validate_model_url(model_name, LIST_OF_URLS)
        self.vector_length = LIST_OF_URLS[model_name]["vector_length"]
        self.model = SentenceTransformer(model_name)

    def get_list_of_urls(self):
        """
            Return list of URLS.
        """
        return self.list_of_urls

    @catch_vector_errors
    def encode(self, text: str) -> List[float]:
        """
            Encode word from transformers.
            This takes the beginning set of tokens and turns them into vectors
            and returns mean pooling of the tokens.
            Args:
                word: string 
        """
        return self.model.encode([text])[0].tolist()
    
    @catch_vector_errors
    def bulk_encode(self, texts: List[str]) -> List[List[float]]:
        """
            Bulk encode words from transformers.
        """
        return self.model.encode(texts).tolist()



from datetime import date
from typing import List
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseTextText2Vec
if is_all_dependency_installed(MODEL_REQUIREMENTS['text-bi-encoder-tfhub-lareqa-qa']):
    import bert
    import numpy as np
    import tensorflow.compat.v2 as tf
    import tensorflow_hub as hub
    from tensorflow.python.framework.errors_impl import NotFoundError
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

LAReQAModelDefinition = ModelDefinition(markdown_filepath='bi_encoders/text_text/tfhub/lareqa_qa')
__doc__ = LAReQAModelDefinition.create_docs()

class LAReQA2Vec(BaseTextText2Vec):
    definition = LAReQAModelDefinition
    def __init__(self, model_url='https://tfhub.dev/google/LAReQA/mBERT_En_En/1', 
    vector_length=512):
        list_of_urls = [
            "https://tfhub.dev/google/LAReQA/mBERT_En_En/1",
            "https://tfhub.dev/google/LAReQA/mBERT_X_X/1",
            "https://tfhub.dev/google/LAReQA/mBERT_X_Y/1",
            "https://tfhub.dev/google/LAReQA/mBERT_X_X_mono/1",
        ]
        self.validate_model_url(model_url, list_of_urls)
        self.model_url = model_url
        self.model = hub.load(self.model_url)
        self.model_name = model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = vector_length
        self.question_encoder = self.model.signatures["query_encoder"]
        self.answer_encoder = self.model.signatures['response_encoder']


    @catch_vector_errors
    def encode_question(self, question: str):
        """
            Encode the question using LAReQA model.
            Example:
            
            >>> from vectorhub.bi_encoders.text_text.tfhub.lareqa_qa import *
            >>> model = LAReQA2Vec()
            >>> model.encode_question("Why?")
        """
        return self.question_encoder(input=tf.constant(np.asarray([question])))["outputs"][0].numpy().tolist()

    @catch_vector_errors
    def bulk_encode_question(self, questions: list):
        """
            Encode questions using LAReQA model.
            Example:
            
            >>> from vectorhub.bi_encoders.text_text.tfhub.lareqa_qa import *
            >>> model = LAReQA2Vec()
            >>> model.encode_question(["Why?", "Who?"])
        """
        return self.question_encoder(input=tf.constant(np.asarray(questions)))["outputs"].numpy().tolist()
    
    @catch_vector_errors
    def encode_answer(self, answer: str, context: str=None):
        """
            Encode answer using LAReQA model.
            Example:
            
            >>> from vectorhub.bi_encoders.text_text.tfhub.lareqa_qa import *
            >>> model = LAReQA2Vec()
            >>> model.encode_answer("Why?")
        """
        if context is None:
            context = answer
        return self.answer_encoder(
            input=tf.constant(np.asarray([answer])),
            context=tf.constant(np.asarray([context])))["outputs"][0].numpy().tolist()

    @catch_vector_errors
    def bulk_encode_answers(self, answers: List[str], contexts: List[str]=None):
        if contexts is None:
            contexts = answers
        return self.answer_encoder(
            input=tf.constant(np.asarray(answers)),
            context=tf.constant(np.asarray(contexts)))["outputs"].numpy().tolist()



from typing import List
from ..base import BaseTextText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from datetime import date

if is_all_dependency_installed(MODEL_REQUIREMENTS['text-bi-encoder-tfhub-use-qa']):
    import bert
    import numpy as np
    import tensorflow as tf
    import tensorflow_hub as hub
    from tensorflow.python.framework.errors_impl import NotFoundError
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

USEQAModelDefinition = ModelDefinition(markdown_filepath='bi_encoders/text_text/tfhub/use_qa')
__doc__ = USEQAModelDefinition.create_docs()

class USEQA2Vec(BaseTextText2Vec):
    definition = USEQAModelDefinition
    def __init__(self):
        self.model_url = "https://tfhub.dev/google/universal-sentence-encoder-qa/3"
        self.model = hub.load(self.model_url)
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 512

    @catch_vector_errors
    def encode_question(self, question: str):
        return self.model.signatures['question_encoder'](tf.constant([question]))['outputs'].numpy().tolist()[0]
    
    @catch_vector_errors
    def bulk_encode_questions(self, questions: List[str]):
        return self.model.signatures['question_encoder'](tf.constant([question]))['outputs'].numpy().tolist()

    @catch_vector_errors
    def encode_answer(self, answer: str, context: str=None):
        if context is None:
            context = answer
        return self.model.signatures['response_encoder'](
            input=tf.constant([answer]),
            context=tf.constant([context]))['outputs'].numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode_answers(self, answers: List[str], contexts: List[str]=None):
        if contexts is None:
            contexts = answers
        return self.model.signatures['response_encoder'](
            input=tf.constant(answers),
            context=tf.constant(contexts))['outputs'].numpy().tolist()



import warnings
import numpy as np
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseText2Vec
from .use import USE2Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-use-multi']):
    import tensorflow as tf
    if hasattr(tf, 'executing_eagerly'):
        if not tf.executing_eagerly():
            tf.compat.v1.enable_eager_execution()
    from tensorflow.python.framework.errors_impl import NotFoundError
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

USEMultiModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/use_multi')

__doc__ = USEMultiModelDefinition.create_docs()

class USEMulti2Vec(USE2Vec):
    definition = USEMultiModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/universal-sentence-encoder-multilingual/3'):
        self.validate_model_url(model_url, list(self.urls.keys()))
        self.init(model_url)
        self.vector_length = 512

    @property
    def urls(self):
        return {
            "https://tfhub.dev/google/universal-sentence-encoder-multilingual/3": {'vector_length': 512},
            "https://tfhub.dev/google/universal-sentence-encoder-multilingual-large/3": {'vector_length': 512}
        }



from datetime import date
from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from ....doc_utils import ModelDefinition

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-albert']):
    from tensorflow.python.framework.errors_impl import NotFoundError
    import tensorflow_hub as hub
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

AlbertModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/albert')
__doc__ = AlbertModelDefinition.create_docs()

class Albert2Vec(BaseText2Vec):
    definition = AlbertModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/tensorflow/albert_en_base/2', max_seq_length: int = 228, normalize: bool = True, 
        preprocessor_url:str ='http://tfhub.dev/tensorflow/albert_en_preprocess/1'):
        self.validate_model_url(model_url, list(self.urls.keys()))
        self.max_seq_length = max_seq_length
        self.normalize = normalize
        self.init(model_url)
        self.init_tokenizer(preprocessor_url)
    
    @property
    def urls(self):
        return {
            'https://tfhub.dev/tensorflow/albert_en_base/1': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/albert_en_xxlarge/1': {'vector_length': 4096},
            'https://tfhub.dev/tensorflow/albert_en_large/1': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/albert_en_xlarge/1': {'vector_length': 2048},
            'https://tfhub.dev/tensorflow/albert_en_base/2': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/albert_en_xxlarge/2': {'vector_length': 4096},
            'https://tfhub.dev/tensorflow/albert_en_large/2': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/albert_en_xlarge/2': {'vector_length': 2048},
        }

    def init_tokenizer(self, preprocessor_url):
        self.preprocessor = hub.KerasLayer(preprocessor_url)

    def init(self, model_url):
        self.model = hub.KerasLayer(model_url)

    @catch_vector_errors
    def encode(self, text: str, pooling_strategy='pooled_output'):
        return self.model(self.preprocessor([text]))[pooling_strategy].numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, texts: list, pooling_strategy='pooled_output'):
        return self.model(self.preprocessor(texts))[pooling_strategy].numpy().tolist()



"""
    Face2Vec Embedding.
"""
from typing import Union
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec

if is_all_dependency_installed("encoders-image-tensorflow-face-detection"):
    import tensorflow as tf
    if hasattr(tf, 'executing_eagerly'):
        if not tf.executing_eagerly():
            tf.compat.v1.enable_eager_execution()
    import appdirs
    import cv2
    import numpy as np
    import requests
    import os
    from mtcnn.mtcnn import MTCNN
    from numpy import asarray
    from PIL import Image
    from keras.models import load_model

FaceNetModelDefinition = ModelDefinition(markdown_filepath="encoders/image/tensorflow/face2vec")
__doc__ = FaceNetModelDefinition.create_docs()

class Face2Vec(BaseImage2Vec):
    definition = FaceNetModelDefinition
    def __init__(self, model_url: str = 'https://drive.google.com/u/0/uc?id=1PZ_6Zsy1Vb0s0JmjEmVd8FS99zoMCiN1&export=download', redownload=True):
        if not os.path.exists(self.model_path) or redownload:
            self._download_model(model_url)
        self.vector_length = self.urls[model_url]["vector_length"]
        self.model = load_model(self.model_path)
    
    def _download_model(self, model_url):
        response = requests.get(model_url)
        if response.status_code != 200:
            raise Exception(response.content)
        with open(self.model_path, 'wb') as f:
            f.write(response.content)

    @property
    def model_path(self):
        return os.path.join(self.cache_dir, 'facenet.h5')

    @property
    def cache_dir(self):
        return appdirs.user_cache_dir()

    @property
    def urls(self):
        """
        A simple dictionary with urls and their vector lengths
        """
        return {'https://drive.google.com/u/0/uc?id=1PZ_6Zsy1Vb0s0JmjEmVd8FS99zoMCiN1&export=download': {'vector_length': 128}}

    def extract_face(self, image_input, reshape_size=None):
        if isinstance(image_input, str):
            pixels = self.read(image_input)
        elif isinstance(image_input, np.ndarray):
            pixels = image_input

        # create the detector, using default weights
        detector = MTCNN()
        # detect faces in the image
        results = detector.detect_faces(pixels)

        # extract the bounding box from the first face
        x1, y1, width, height = results[0]['box']
        # bug fix
        x1, y1 = abs(x1), abs(y1)
        x2, y2 = x1 + width, y1 + height

        # extract the face
        face = pixels[y1:y2, x1:x2]

        # resize pixels to the model size
        image = Image.fromarray(face)
        if reshape_size is not None:
            image = image.resize(reshape_size)
        face_array = asarray(image)
        return face_array
    
    def show_face_landmarks(self, image_filename: str):
        """
        Show face landmarks
        """
        detector = MTCNN()

        # image = cv2.cvtColor(cv2.imread("rose.jpeg"), cv2.COLOR_BGR2RGB)
        image = self.extract_face(image_filename)
        result = detector.detect_faces(image)
        bounding_box = result[0]['box']
        keypoints = result[0]['keypoints']

        cv2.rectangle(image,
        (bounding_box[0], bounding_box[1]),
        (bounding_box[0]+bounding_box[2], bounding_box[1] + bounding_box[3]),
        (0,155,255), 2)
        cv2.circle(image,(keypoints['left_eye']), 2, (0,155,255), 2)
        cv2.circle(image,(keypoints['right_eye']), 2, (0,155,255), 2)
        cv2.circle(image,(keypoints['nose']), 2, (0,155,255), 2)
        cv2.circle(image,(keypoints['mouth_left']), 2, (0,155,255), 2)
        cv2.circle(image,(keypoints['mouth_right']), 2, (0,155,255), 2)
        return plt.imshow(image)

    def standardise_image(self, face_pixels):
        """
        Standardise the image for face pixels.
        """
        face_pixels = face_pixels.astype('float32')
        # standardize pixel values across channels (global)
        mean, std = face_pixels.mean(), face_pixels.std()
        face_pixels = (face_pixels - mean) / std
        return tf.expand_dims(face_pixels, axis=0)

    @catch_vector_errors
    def encode(self, image):
        if isinstance(image, (np.ndarray, str)):
            image = self.standardise_image(self.extract_face(image, reshape_size=(160, 160)))
        return self.model.predict([image]).tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, images):
        """
            Bulk encode. Chunk size should be specified outside of the images.
        """
        # TODO: Change from list comprehension to properly read
        return [self.encode(x) for x in images]



"""
    Vector AI's deployed model. The purpose of this model is to allow developers to easily build encodings and see for themselves
    how the embedding works. These models are selected to work out-of-the-box after testing for their success on our end.

    To get access to Vector AI, we need to use 

    Example:

        >>> from vectorhub.text.encoder.vectorai import ViText2Vec
        >>> model = ViText2Vec(username, api_key)
        >>> model.encode("sample.jpg")
    
"""

import io
import base64
import requests
from ....base import catch_vector_errors

class ViImage2Vec:
    def __init__(self, username, api_key, url=None, collection_name="base"):
        """
            Request for a username and API key from gh.vctr.ai
        """
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"
        self.collection_name = collection_name
        self._name = "default"
    
    @catch_vector_errors
    def encode(self, image):
        return requests.get(
            url="{}/collection/encode_image".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": self.collection_name,
                "image_url": image,
            },
        ).json()

    @property
    def __name__(self):
        if self._name is None:
            return "deployed_image"
        return self._name

    @__name__.setter
    def __name__(self, value):
        self._name = value


import warnings
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseText2Vec
from datetime import date
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-use']):
    import tensorflow_hub as hub
    import tensorflow.compat.v1 as tf
    import numpy as np

ElmoModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/elmo.md')

__doc__ = ElmoModelDefinition.create_docs()

class Elmo2Vec(BaseText2Vec):
    definition = ElmoModelDefinition
    def __init__(self, model_url: str="https://tfhub.dev/google/elmo/3", trainable_model=True):
        warnings.warn("We are disabling TF2 eager execution to run this. This may conflict with other models. If you need + \
            other models., try to use a fresh environment or a new virtual machine.")
        tf.disable_eager_execution()
        self.model = hub.Module(model_url, trainable=trainable_model)
        self.vector_length = 1024
    
    @property
    def urls(self):
        return {
           "https://tfhub.dev/google/elmo/3": {'vector_length': 1024}
        }

    @catch_vector_errors
    def encode(self, text, output_layer: str="elmo"):
        """
        The output layer can be one of the following:
        lstm_outputs1: the first LSTM hidden state with shape [batch_size, max_length, 1024].
        lstm_outputs2: the second LSTM hidden state with shape [batch_size, max_length, 1024].
        elmo: the weighted sum of the 3 layers, where the weights are trainable. This tensor has shape [batch_size, max_length, 1024]
        default: a fixed mean-pooling of all contextualized word representations with shape [batch_size, 1024].
        Note: The output layer word_emb is character-based and is not supported by VectorHub.
        """
        sess = tf.Session()
        init = tf.global_variables_initializer()
        sess.run(init)
        if output_layer != 'default':
            vector = self.model(
                [text],
                signature="default",
                as_dict=True)[output_layer].eval(session=sess)[0][0].tolist()
        else:
            vector = self.model(
                [text],
                signature="default",
                as_dict=True)[output_layer].eval(session=sess)[0].tolist()
        sess.close()
        return vector

    @catch_vector_errors
    def bulk_encode(self, texts, output_layer: str="elmo"):
        """
        The output layer can be one of the following:
            lstm_outputs1: the first LSTM hidden state with shape [batch_size, max_length, 1024].
            lstm_outputs2: the second LSTM hidden state with shape [batch_size, max_length, 1024].
            elmo: the weighted sum of the 3 layers, where the weights are trainable. This tensor has shape [batch_size, max_length, 1024]
            default: a fixed mean-pooling of all contextualized word representations with shape [batch_size, 1024].
        Note: The output layer word_emb is character-based and is not supported by VectorHub.
        """
        sess = tf.Session()
        init = tf.global_variables_initializer()
        sess.run(init)
        vectors = self.model(
            texts,
            signature="default",
            as_dict=True)[output_layer].eval(session=sess).tolist()
        sess.close()
        if output_layer == 'default':
            return vectors
        else:
            return [x[0] for x in vectors]



from datetime import date
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-tfhub-speech_embedding']):
    import tensorflow as tf
    import tensorflow_hub as hub
    import traceback

from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ..base import BaseAudio2Vec

SpeechEmbeddingModelDefinition = ModelDefinition(markdown_filepath="encoders/audio/tfhub/speech_embedding.md")

__doc__ = SpeechEmbeddingModelDefinition.create_docs()

class SpeechEmbedding2Vec(BaseAudio2Vec):
    definition = SpeechEmbeddingModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/speech_embedding/1', signature: str = 'default'):
        self.model_url = model_url
        self.signature = signature
        self.model = hub.load(self.model_url).signatures[self.signature]
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 96

    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/speech_embedding/1': {'vector_length': 96}
        }

    @catch_vector_errors
    def encode(self, audio, vector_operation='mean'):
        """
        Encode the vector.
        Example:

        >>> from vectorhub.encoders.audio import SpeechEmbedding2Vec
        >>> encoder = SpeechEmbedding2Vec()
        >>> encoder.encode(...)
        """
        if isinstance(audio, str):
            audio = self.read(audio)
        return self._vector_operation(self.model(tf.constant([audio]))[self.signature][0], vector_operation=vector_operation)[0]

    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        audios = [self.read(audio) if isinstance(audio, str) else audio for audio in audios]
        return self._vector_operation(self.model(tf.constant(audios))[self.signature][0], vector_operation=vector_operation)



from datetime import date
from typing import List
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseImage2Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-image-tfhub-resnet']):
    import tensorflow as tf
    import tensorflow_hub as hub

ResNetModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/resnet')

__doc__ = ResNetModelDefinition.create_docs()

class ResnetV12Vec(BaseImage2Vec):
    definition = ResNetModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/imagenet/resnet_v1_50/feature_vector/4'):
        list_of_urls = [
            # 50 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_50/feature_vector/4',

            # 101 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_101/feature_vector/4',

            # 152 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_152/feature_vector/4',
        ]
        self.validate_model_url(model_url, list_of_urls)
        self.init(model_url)
        self.vector_length = 2048

    @property
    def urls(self):
        return {
            # 50 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_50/feature_vector/4':{'vector_length': 2048},

            # 101 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_101/feature_vector/4':{'vector_length': 2048},

            # 152 layers
            'https://tfhub.dev/google/imagenet/resnet_v1_152/feature_vector/4':{'vector_length': 2048},
        }


    def init(self, model_url: str):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.model = hub.load(self.model_url)

    @catch_vector_errors
    def encode(self, image):
        if isinstance(image, str):
            image = self.read(image)
        return self.model([image]).numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, images: List[str]):
        return [self.encode(x) for x in images]



"""
    Utilities for importing libraries.
"""
import sys
import warnings
import pkg_resources
import json
from importlib import import_module, invalidate_caches

def get_package_requirements(requirement_type: str):
    """
        Load in extra_requirements.json from the package
    """
    requirements = {
        "numpy": ["core"],
        "requests": ["core"],
        "PyYAML": ["core"],
        "pytest": ["test"],
        "sphinx-rtd-theme>=0.5.0": ["test"],
        "imageio": ["encoders-image", "encoders-image-tfhub"],
        "scikit-image":  ["encoders-image", "encoders-image-tfhub"],
        "soundfile": ["encoders-audio-tfhub"],
        "librosa": ["audio-encoder", "encoders-audio-tfhub"],
        "tensorflow": ["encoders-text-tfhub", "encoders-audio-tfhub", "encoders-image-tfhub", "encoders-text-tf-transformers", 
            "encoders-text-tfhub-windows", "encoders-image-tensorflow-face-detection"],
        "tensorflow_hub": ["encoders-text-tfhub", "encoders-audio-tfhub", "encoders-image-tfhub", "encoders-text-tfhub-windows"],
        "tensorflow_text": ["encoders-text-tfhub"],
        "tf-models-official": ["encoders-text-tfhub", "encoders-text-tfhub-windows"],
        "bert-for-tf2": ["encoders-text-tfhub", "encoders-text-tfhub-windows"],
        "sentence-transformers": ["encoders-text-sentence-transformers"],
        "torch>=1.6.0": ["encoders-audio-pytorch", "encoders-text-torch-transformers", "encoders-text-sentence-transformers", "encoders-image-fastai"],
        "fairseq": ["encoders-audio-pytorch"],
        "transformers": ["encoders-text-torch-transformers", "encoders-text-tf-transformers"],
        "moviepy": ["encoders-video"],
        "opencv-python": ["encoders-video", "encoders-image-tensorflow-face-detection"],
        "appdirs": ["encoders-image-tensorflow-face-detection"],
        "fastai==2.1.8": ["encoders-image-fastai"],
        "mtcnn": ["encoders-image-tensorflow-face-detection"],
        "Pillow": ["encoders-image-tensorflow-face-detection"]
    }
    dependencies = []
    for k, v in requirements.items():
        if requirement_type in v:
            dependencies.append(k) 
    return dependencies

def is_dependency_installed(dependency: str):
    """
        Returns True if the dependency is installed else False.
    """
    IS_INSTALLED = True
    try:
        pkg_resources.get_distribution(dependency)
    except pkg_resources.ContextualVersionConflict:
        IS_INSTALLED = True
    except:
        IS_INSTALLED = False
    return IS_INSTALLED

def is_all_dependency_installed(requirement_type: str, raise_warning=True):
    """
        Returns True/False if the dependency is isntalled
        Args:
            requirement_type: The type of requirement. This can be found in the values in extra_requirements.json
            raise_warning: Raise warning if True
    """
    IS_ALL_INSTALLED = True
    requirements = get_package_requirements(requirement_type)
    for r in requirements:
        if not is_dependency_installed(r):
            if raise_warning:
                warnings.warn(f"You are missing {r} dependency for this submodule. Run `pip install vectorhub[{requirement_type}]`")
            IS_ALL_INSTALLED = False
    return IS_ALL_INSTALLED



from .bit import * 
from .bit_medium import *
from .inceptionv1 import *
from .inceptionv2 import *
from .inceptionv3 import *
from .inception_resnet import *
from .mobilenet import *
from .mobilenetv2 import *
from .resnet import *
from .resnetv2 import *



import numpy as np
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec
if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub
    import io
    import imageio
    import numpy as np
    import matplotlib.pyplot as plt
    from urllib.request import urlopen, Request
    from urllib.parse import quote
    from skimage import transform

MobileNetModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/mobilenet')

__doc__ = MobileNetModelDefinition.create_docs()

class MobileNetV12Vec(BaseImage2Vec):
    definition = MobileNetModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/imagenet/mobilenet_v1_100_224/feature_vector/4', resize_mode: str='symmetric'):
        list_of_urls = {
            # 100 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_224/feature_vector/4': {"vector_length":1024, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_192/feature_vector/4': {"vector_length":1024, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_160/feature_vector/4': {"vector_length":1024, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_128/feature_vector/4': {"vector_length":1024, "image_dimensions":128},

            # 75 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_224/feature_vector/4': {"vector_length":768, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_192/feature_vector/4': {"vector_length":768, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_160/feature_vector/4': {"vector_length":768, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_128/feature_vector/4': {"vector_length":768, "image_dimensions":128},

            # 50 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_224/feature_vector/4': {"vector_length":512, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_192/feature_vector/4': {"vector_length":512, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_160/feature_vector/4': {"vector_length":512, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_128/feature_vector/4': {"vector_length":512, "image_dimensions":128},

            # 25 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_224/feature_vector/4': {"vector_length":256, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_192/feature_vector/4': {"vector_length":256, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_160/feature_vector/4': {"vector_length":256, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_128/feature_vector/4': {"vector_length":256, "image_dimensions":128},
        }
        self.validate_model_url(model_url, list_of_urls)
        self.vector_length = list_of_urls[model_url]["vector_length"]
        self.image_dimensions = list_of_urls[model_url]["image_dimensions"]
        self.init(model_url)
        self.resize_mode = resize_mode

    @property
    def urls(self):
        return {
            # 100 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_224/feature_vector/4': {"vector_length":1024, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_192/feature_vector/4': {"vector_length":1024, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_160/feature_vector/4': {"vector_length":1024, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_100_128/feature_vector/4': {"vector_length":1024, "image_dimensions":128},

            # 75 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_224/feature_vector/4': {"vector_length":768, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_192/feature_vector/4': {"vector_length":768, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_160/feature_vector/4': {"vector_length":768, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_075_128/feature_vector/4': {"vector_length":768, "image_dimensions":128},

            # 50 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_224/feature_vector/4': {"vector_length":512, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_192/feature_vector/4': {"vector_length":512, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_160/feature_vector/4': {"vector_length":512, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_050_128/feature_vector/4': {"vector_length":512, "image_dimensions":128},

            # 25 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_224/feature_vector/4': {"vector_length":256, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_192/feature_vector/4': {"vector_length":256, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_160/feature_vector/4': {"vector_length":256, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v1_025_128/feature_vector/4': {"vector_length":256, "image_dimensions":128},
        }


    def init(self, model_url: str):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.model = tf.keras.Sequential([
            hub.KerasLayer(self.model_url, trainable=False)
        ])
        self.model.build([None, self.image_dimensions, self.image_dimensions, 3])
    
    @catch_vector_errors
    def encode(self, image):
        if isinstance(image, str):
            image = self.image_resize(self.read(image), self.image_dimensions, self.image_dimensions,
            resize_mode=self.resize_mode)[np.newaxis, ...]
        return self.model(image).numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, images):
        """
            Bulk encode. Chunk size should be specified outside of the images.
        """
        # TODO: Change from list comprehension to properly read in bulk
        return [self.encode(x) for x in images]



from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from datetime import date
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-bert']):
    import tensorflow as tf
    if hasattr(tf, 'executing_eagerly'):
        if not tf.executing_eagerly():
            tf.compat.v1.enable_eager_execution()
    import tensorflow_hub as hub
    import bert
    import numpy as np

BertModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/bert')

__doc__ = BertModelDefinition.create_docs()

class Bert2Vec(BaseText2Vec):
    definition = BertModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/3', 
    max_seq_length: int = 64, normalize: bool = True):
        list_of_urls = [
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/2',
            'https://tfhub.dev/tensorflow/bert_en_wwm_cased_L-24_H-1024_A-16/2',
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/2',
            'https://tfhub.dev/tensorflow/bert_en_wwm_uncased_L-24_H-1024_A-16/2',
            'https://tfhub.dev/tensorflow/bert_en_cased_L-24_H-1024_A-16/2',
            'https://tfhub.dev/tensorflow/bert_en_cased_L-12_H-768_A-12/2',
            'https://tfhub.dev/tensorflow/bert_zh_L-12_H-768_A-12/2',
            'https://tfhub.dev/tensorflow/bert_multi_cased_L-12_H-768_A-12/2',

            'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/3',
            'https://tfhub.dev/tensorflow/bert_en_wwm_cased_L-24_H-1024_A-16/3',
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/3',
            'https://tfhub.dev/tensorflow/bert_en_wwm_uncased_L-24_H-1024_A-16/3',
            'https://tfhub.dev/tensorflow/bert_en_cased_L-24_H-1024_A-16/3',
            'https://tfhub.dev/tensorflow/bert_en_cased_L-12_H-768_A-12/3',
            'https://tfhub.dev/tensorflow/bert_zh_L-12_H-768_A-12/3',
            'https://tfhub.dev/tensorflow/bert_multi_cased_L-12_H-768_A-12/3',
        ]
        self.validate_model_url(model_url, list_of_urls)
        self.max_seq_length = max_seq_length
        self.normalize = normalize
        self.model_input_type = "dict"
        self.init(model_url)
        self.tokenizer = self.init_tokenizer()

    @property
    def urls(self):
        {
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/2': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_wwm_cased_L-24_H-1024_A-16/2': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/2': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_en_wwm_uncased_L-24_H-1024_A-16/2': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_cased_L-24_H-1024_A-16/2': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_cased_L-12_H-768_A-12/2': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_zh_L-12_H-768_A-12/2': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_multi_cased_L-12_H-768_A-12/2': {'vector_length': 768},

            'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/3': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_wwm_cased_L-24_H-1024_A-16/3': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/3': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_en_wwm_uncased_L-24_H-1024_A-16/3': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_cased_L-24_H-1024_A-16/3': {'vector_length': 1024},
            'https://tfhub.dev/tensorflow/bert_en_cased_L-12_H-768_A-12/3': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_zh_L-12_H-768_A-12/3': {'vector_length': 768},
            'https://tfhub.dev/tensorflow/bert_multi_cased_L-12_H-768_A-12/3': {'vector_length': 768},
        }

    def init(self, model_url: str):
        self.model = hub.KerasLayer(model_url)
        input_word_ids = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32)
        input_mask = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32)
        input_type_ids = tf.keras.layers.Input(shape=(self.max_seq_length,), dtype=tf.int32)
        try:
            self.model(dict(input_word_ids=input_word_ids, input_mask=input_mask, input_type_ids=input_type_ids))
        except ValueError:
            self.model_input_type = "list"
            self.model([input_word_ids, input_mask, input_type_ids])

    def init_tokenizer(self):
        self.vocab_file = self.model.resolved_object.vocab_file.asset_path.numpy()
        self.do_lower_case = self.model.resolved_object.do_lower_case.numpy()
        return bert.bert_tokenization.FullTokenizer(self.vocab_file, self.do_lower_case)

    def process(self, input_strings: str):
        input_ids_all, input_mask_all, input_type_ids_all = [], [], []
        if isinstance(input_strings, str):
            input_strings = [input_strings]
        for input_string in input_strings:
            # Tokenize input.
            input_tokens = ["[CLS]"] + \
                self.tokenizer.tokenize(input_string) + ["[SEP]"]
            input_ids = self.tokenizer.convert_tokens_to_ids(input_tokens)
            sequence_length = min(len(input_ids), self.max_seq_length)

            # Padding or truncation.
            if len(input_ids) >= self.max_seq_length:
                input_ids = input_ids[:self.max_seq_length]
            else:
                input_ids = input_ids + [0] * \
                    (self.max_seq_length - len(input_ids))

            input_mask = [1] * sequence_length + [0] * \
                (self.max_seq_length - sequence_length)

            input_ids_all.append(input_ids)
            input_mask_all.append(input_mask)
            input_type_ids_all.append([0] * self.max_seq_length)

        return np.array(input_ids_all), np.array(input_mask_all),  np.array(input_type_ids_all)

    @catch_vector_errors
    def encode(self, text: str, pooling_strategy='pooled_output'):
        input_ids, input_mask, input_type_ids = self.process(text)
        if self.model_input_type == "list":
            return self.model([
                tf.convert_to_tensor(input_ids, tf.int32, name="input_word_ids"), 
                tf.convert_to_tensor(input_mask, tf.int32, name="input_mask"), 
                tf.convert_to_tensor(input_type_ids, tf.int32, name="input_type_ids")
            ])[0].numpy().tolist()[0]
        else:
            return self.model({
                "input_word_ids": tf.convert_to_tensor(input_ids, tf.int32, name="input_word_ids"), 
                "input_mask": tf.convert_to_tensor(input_mask, tf.int32, name="input_mask"), 
                "input_type_ids": tf.convert_to_tensor(input_type_ids, tf.int32, name="input_type_ids")
            })[pooling_strategy].numpy().tolist()[0]


    @catch_vector_errors
    def bulk_encode(self, texts: list, pooling_strategy='pooled_output'):
        input_ids, input_mask, input_type_ids = self.process(texts)
        if self.model_input_type == "list":
            return self.model([
                tf.convert_to_tensor(input_ids, tf.int32, name="input_word_ids"), 
                tf.convert_to_tensor(input_mask, tf.int32, name="input_mask"), 
                tf.convert_to_tensor(input_type_ids, tf.int32, name="input_type_ids")
            ])[0].numpy().tolist()
        else:
            return self.model({
                "input_word_ids": tf.convert_to_tensor(input_ids, tf.int32, name="input_word_ids"), 
                "input_mask": tf.convert_to_tensor(input_mask, tf.int32, name="input_mask"), 
                "input_type_ids": tf.convert_to_tensor(input_type_ids, tf.int32, name="input_type_ids")
            })[pooling_strategy].numpy().tolist()



from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec
from .bit import BitSmall2Vec

if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub
    import traceback

BITMediumModelDefinition = ModelDefinition(markdown_filepath="encoders/image/tfhub/bit_medium")
__doc__ = BITMediumModelDefinition.create_docs()

class BitMedium2Vec(BitSmall2Vec):
    definition = BITMediumModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/bit/m-r50x1/1'):
        list_of_urls = {
            'https://tfhub.dev/google/bit/m-r50x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/m-r50x3/1': {"vector_length":6144}, # 6144 output shape
            'https://tfhub.dev/google/bit/m-r101x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/m-r101x3/1': {"vector_length":6144},  # 6144 output shape
            'https://tfhub.dev/google/bit/m-r152x4/1': {"vector_length":8192},  # 8192 output shape
        }
        self.validate_model_url(model_url, list_of_urls)
        self.init(model_url)
        self.vector_length = list_of_urls[model_url]["vector_length"]
    
    @property
    def urls(self):
        """
        A simple dictionary with urls and their vector lengths
        """
        return {
            'https://tfhub.dev/google/bit/m-r50x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/m-r50x3/1': {"vector_length":6144}, # 6144 output shape
            'https://tfhub.dev/google/bit/m-r101x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/m-r101x3/1': {"vector_length":6144},  # 6144 output shape
            'https://tfhub.dev/google/bit/m-r152x4/1': {"vector_length":8192},  # 8192 output shape
        }



import numpy as np
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec
from .mobilenet import MobileNetV12Vec
if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub

MobileNetV2ModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/mobilenetv2')

__doc__ = MobileNetV2ModelDefinition.create_docs()

class MobileNetV22Vec(MobileNetV12Vec):
    definition = MobileNetV2ModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/imagenet/mobilenet_v2_140_224/feature_vector/4', resize_mode:str="symmetric"):
        list_of_urls = {
            # 140 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_140_224/feature_vector/4': {"vector_length":1792, "image_dimensions":224},

            # 130 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_130_224/feature_vector/4': {"vector_length":1664, "image_dimensions":224},

            # 100 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 75 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 50 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 35 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},
        }
        self.validate_model_url(model_url, list_of_urls)
        self.vector_length = list_of_urls[model_url]["vector_length"]
        self.image_dimensions = list_of_urls[model_url]["image_dimensions"]
        self.init(model_url)
        self.resize_mode = resize_mode
    
    @property
    def urls(self):
        """
        URLs and their metadata
        """
        return {
            # 140 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_140_224/feature_vector/4': {"vector_length":1792, "image_dimensions":224},

            # 130 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_130_224/feature_vector/4': {"vector_length":1664, "image_dimensions":224},

            # 100 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_100_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 75 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_075_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 50 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_050_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},

            # 35 depth
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_224/feature_vector/4': {"vector_length":1280, "image_dimensions":224},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_192/feature_vector/4': {"vector_length":1280, "image_dimensions":192},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_160/feature_vector/4': {"vector_length":1280, "image_dimensions":160},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_128/feature_vector/4': {"vector_length":1280, "image_dimensions":128},
            'https://tfhub.dev/google/imagenet/mobilenet_v2_035_96/feature_vector/4': {"vector_length":1280, "image_dimensions":96},
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import codecs
import sys
import json
import re
from setuptools import setup,find_packages
from collections import defaultdict
from typing import List, Dict
from pathlib import Path


def get_extra_requires(path, add_all=True):
    if '.json' in path:
        try:
            requirements_dict = json.load(open(path, 'r'))
            return dependency_to_requirement(requirements_dict)
        except FileNotFoundError:
            print(f"{path} not found")
            return {}

def dependency_to_requirement(requirements_dict: Dict, add_all=True, add_single_package=True):
    """
        Invert the index from dependency to requirement.
    """
    all_requirements = defaultdict(set)
    for library, dependency in requirements_dict.items():
        for d in dependency:
            all_requirements[d].add(library)
    if add_single_package:
        for k in requirements_dict.keys():
            all_requirements[k] = {k}
    if add_all:
        all_requirements['all'] = set(v for v in requirements_dict.keys())
    return all_requirements

all_deps = get_extra_requires('extra_requirements.json')

# Additional files to include - adding model cards
# package_data = [str(x) for x in Path('vectorhub').rglob('*.md')]
package_data = [str(x) for x in list(Path('vectorhub').rglob("*.md"))]

# Also add the extra_requirements.json file
package_data.append('extra_requirements.json')

version = '1.0.8'

if 'IS_VECTORHUB_NIGHTLY' in os.environ.keys():
    from datetime import datetime
    name = 'vectorhub-nightly'
    version = version + '.' + datetime.today().date().__str__().replace('-', '.') 
else:
    name = 'vectorhub'

setup(
    name=name,
    version=version,
    author="OnSearch Pty Ltd",
    author_email="dev@vctr.ai",
    package_data={'vectorhub': package_data},
    include_package_data=True,
    # data_files=[('vectorhub', package_data)], # puts the markdown files in a new directory - not what we want
    description="One liner to encode data into vectors with state-of-the-art models using tensorflow, pytorch and other open source libraries. Word2Vec, Image2Vec, BERT, etc",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="vector, embeddings, machinelearning, ai, artificialintelligence, nlp, tensorflow, pytorch, nearestneighbors, search, analytics, clustering, dimensionalityreduction",
    url="https://github.com/vector-ai/vectorhub",
    license="Apache",
    packages=find_packages(exclude=["tests*"]),
    python_requires=">=3",
    install_requires=list(all_deps['core'].union(all_deps['perf'])),
    extras_require=all_deps,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
        "Topic :: Multimedia :: Video :: Conversion",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseImage2Vec
from .resnet import ResnetV12Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-image-tfhub-resnet']):
    import tensorflow as tf
    import tensorflow_hub as hub

ResNetV2ModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/resnetv2')
__doc__ = ResNetV2ModelDefinition.create_docs()

class ResnetV22Vec(ResnetV12Vec):
    definition = ResNetV2ModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/imagenet/resnet_v2_50/feature_vector/4'):
        list_of_urls = [
            # 50 layers
            'https://tfhub.dev/google/imagenet/resnet_v2_50/feature_vector/4',

            # 101 layers
            'https://tfhub.dev/google/imagenet/resnet_v2_101/feature_vector/4',

            # 152 layers
            'https://tfhub.dev/google/imagenet/resnet_v2_152/feature_vector/4'
        ]
        self.validate_model_url(model_url, list_of_urls)
        self.init(model_url)
        self.vector_length = 2048

    @property
    def urls(self):
        return {
            'https://tfhub.dev/google/imagenet/resnet_v2_50/feature_vector/4': {'vector_length': 2048}, 

            # 101 layers
            'https://tfhub.dev/google/imagenet/resnet_v2_101/feature_vector/4':{'vector_length': 2048}, 

            # 152 layers
            'https://tfhub.dev/google/imagenet/resnet_v2_152/feature_vector/4':{'vector_length': 2048}, 
        }



from typing import List
from ..base import BaseTextText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from datetime import date
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-sentence-transformers']):
    from sentence_transformers import SentenceTransformer

DistilRobertaQAModelDefinition = ModelDefinition(markdown_filepath='bi_encoders/text_text/sentence_transformers/distilroberta_qa.md')

__doc__ = DistilRobertaQAModelDefinition.create_docs()


class DistilRobertaQA2Vec(BaseTextText2Vec):
    definition = DistilRobertaQAModelDefinition
    def __init__(self):
        self.model = SentenceTransformer('distilroberta-base-msmarco-v1')
        self.vector_length = 768
    
    @catch_vector_errors
    def encode_question(self, question: str):
        return self.model.encode(["[QRY] "+ question])[0].tolist()

    @catch_vector_errors
    def bulk_encode_question(self, questions: list):
        return [self.encode(q) for q in questions]
    
    @catch_vector_errors
    def encode_answer(self, answer: str, context: str=None):
        return self.model.encode(["[DOC] "+ answer])[0].tolist()

    @catch_vector_errors
    def bulk_encode_answers(self, answers: List[str]):
        return [self.encode(a) for a in answers]



from datetime import date
from ....base import catch_vector_errors 
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseAudio2Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-audio-pytorch-fairseq']):
    import torch
    import numpy as np
    from fairseq.models.wav2vec import Wav2VecModel

WavModelDefinition = ModelDefinition(markdown_filepath='encoders/audio/pytorch/wav2vec')

class Wav2Vec(BaseAudio2Vec):
    definition = WavModelDefinition
    def __init__(self, model_url: str = 'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small.pt'):
        self.list_of_urls = [
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_10m.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_100h.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_960h.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox_10m.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox_100h.pt',
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec2_vox_960h.pt',
        ]
        self.validate_model_url(model_url, self.list_of_urls)
        self.init(model_url)
        self.vector_length = 512

    def init(self, model_url: str):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://dl.fbaipublicfiles.com/fairseq/', '').replace('/', '_')
        torch_model = torch.hub.load_state_dict_from_url(self.model_url)
        self.model = Wav2VecModel.build_model(torch_model['args'], task=None)

    @property
    def urls(self):
        return {
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_10m.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_100h.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_small_960h.pt': {}, 
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox_10m.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec_vox_100h.pt': {},
            'https://dl.fbaipublicfiles.com/fairseq/wav2vec/wav2vec2_vox_960h.pt': {},
        }

    @catch_vector_errors
    def encode(self, audio, vector_operation='mean'):
        """
        Example:

            >>> from vectorhub.encoders.audio import Wav2Vec
            >>> encoder = Wav2Vec()
            >>> encoder.encode("...")
        """
        if isinstance(audio, str):
            audio = self.read(audio)
        return self._vector_operation(self.model.feature_extractor(torch.from_numpy(np.array([audio]))).detach().numpy().tolist()[0], vector_operation=vector_operation, axis=1)

    @catch_vector_errors
    def bulk_encode(self, audios, vector_operation='mean'):
        return [self.encode(audio, vector_operation=vector_operation) for audio in audios]



from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseText2Vec
from datetime import date
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-use']):
    import tensorflow_hub as hub
    import tensorflow as tf
    if hasattr(tf, 'executing_eagerly'):
        if not tf.executing_eagerly():
            tf.compat.v1.enable_eager_execution()

USEModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/use.md')

__doc__ = USEModelDefinition.create_docs()

class USE2Vec(BaseText2Vec):
    definition = USEModelDefinition
    # or layer19
    def __init__(self, model_url: str = 'https://tfhub.dev/google/universal-sentence-encoder/4'):
        self.validate_model_url(model_url, list(self.urls.keys()))
        self.init(model_url)
        self.vector_length = 512

    @property
    def urls(self):
        return {
            "https://tfhub.dev/google/universal-sentence-encoder/4": {'vector_length': 512},
            "https://tfhub.dev/google/universal-sentence-encoder-large/5": {'vector_length': 512}
        }

    def init(self, model_url: str):
        self.model_url = model_url
        self.model = hub.load(self.model_url)
        self.model_name = model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')


    @catch_vector_errors
    def encode(self, text):
        return self.model([text]).numpy().tolist()[0]

    # can consider compress in the future
    @catch_vector_errors
    def bulk_encode(self, texts, threads=10, chunks=100):
        return [i for c in self.chunk(texts, chunks) for i in self.model(c).numpy().tolist()]
